import ast
import datetime
import json
import sys
import urllib.parse
import arcgis.gis
import re

def _camelCase_to_underscore(name):
    """PEP8ify name"""
    name = name.replace(" ", "_")
    if '_' in name:
        return name.lower()
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

def _underscore_to_camelcase(name):
    def camelcase(): 
        yield str.lower
        while True:
            yield str.capitalize

    c = camelcase()
    return "".join(next(c)(x) if x else '_' for x in name.split("_"))

class WorkflowManagerAdmin:
    """
    Represents a a series of CRUD functions for Workflow Manager Items

    ===============     ====================================================================
    **Argument**        **Description**
    ---------------     --------------------------------------------------------------------
    gis                 Optional GIS. The connection to the Enterprise.
    ===============     ====================================================================
    """

    def __init__(self, gis):
        self._gis = gis
        if self._gis.users.me is None:
            raise ValueError("An authenticated `GIS` is required.")
        if not any(prov.itemid == '50a5f00bcc574358b15eab0e2bdadf39' for prov in self._gis.users.me.provisions):
            raise ValueError("No Workflow Manager license is available for the current user")
        self._url = self._wmx_server_url[0]
        if self._url is None:
            raise ValueError("No WorkflowManager Registered with your Organization")
        if not any(prov.itemid == '50a5f00bcc574358b15eab0e2bdadf39' for prov in self._gis.users.me.provisions):
            raise ValueError("No Workflow Manager license is available for the current user")

    @property
    def _wmx_server_url(self):
        """locates the WMX server"""
        baseurl = self._gis._portal.resturl
        res = self._gis._con.get(f"{baseurl}/portals/self/servers", {'f': 'json'})
        for s in res['servers']:
            server_functions = [x.strip() for x in s.get("serverFunction", "").lower().split(",")]
            if 'workflowmanager' in server_functions:
                self._url = s.get("url", None)
                self._private_url = s.get("adminUrl", None)
                if self._url is None:
                    raise RuntimeError("Cannot find a WorkflowManager Server")
                self._url += f"/workflow"
                self._private_url += f"/workflow"
                return self._url, self._private_url
        return None

    def create_item(self, name) -> tuple:
        """
        Creates a `Workflow Manager` schema that stores all the configuration 
        information and location data in the data store on Portal. This can
        be run by any user assigned to the administrator role in Portal.
        
        For users that do not belong to the administrator role, the 
        following privileges are required to run Create Workflow Item:
        
        ==================  =========================================================
        **Argument**        **Description**
        ------------------  ---------------------------------------------------------
        name                Required String. The name of the new schema.
        ==================  =========================================================
        
        :returns: string (item_id)
        """
        url = '{base}/admin/createWorkflowItem?token={token}&name={name}'.format(base=self._url,
                                                                                 token=self._gis._con.token, name=name)
        params = {
            'name': name
        }
        return_obj = json.loads(
            self._gis._con.post(url, params=params, try_json=False, add_token=False, json_encode=False,
                                post_json=True))['itemId']
        if 'error' in return_obj:
            self._gis._con._handle_json_error(return_obj['error'], 0)
        elif 'success' in return_obj:
            return res['success']
        return return_obj


class JobManager:
    """
    Represents a helper class for workflow manager jobs

    ===============     ====================================================================
    **Argument**        **Description**
    ---------------     --------------------------------------------------------------------
    item                The Workflow Manager Item
    ===============     ====================================================================

    """

    def __init__(self, item):
        """initializer"""
        if item is None:
            raise ValueError("Item cannot be None")
        self._item = item
        self._gis = item._gis
        if self._gis.users.me is None:
            raise ValueError("An authenticated `GIS` is required.")

        self._url = self._wmx_server_url[0]
        if self._url is None:
            raise ValueError("No WorkflowManager Registered with your Organization")
        if not any(prov.itemid == '50a5f00bcc574358b15eab0e2bdadf39' for prov in self._gis.users.me.provisions):
            raise ValueError("No Workflow Manager license is available for the current user")
        
    def _handle_error(self, info):
        """Basic error handler - separated into a function to allow for expansion in future releases"""
        error_class = info[0]
        error_text = info[1]
        raise Exception(error_text)

    @property
    def _wmx_server_url(self):
        """locates the WMX server"""
        baseurl = self._gis._portal.resturl
        res = self._gis._con.get(f"{baseurl}/portals/self/servers", {'f': 'json'})
        for s in res['servers']:
            server_functions = [x.strip() for x in s.get("serverFunction", "").lower().split(",")]
            if 'workflowmanager' in server_functions:
                self._url = s.get("url", None)
                self._private_url = s.get("adminUrl", None)
                if self._url is None:
                    raise RuntimeError("Cannot find a WorkflowManager Server")
                self._url += f"/workflow/{self._item.id}"
                self._private_url += f"/workflow/{self._item.id}"
                return self._url, self._private_url
        return None

    def close(self, job_ids):
        """
        Closes a single or multiple jobs with specific JobIDs

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        job_ids             A list of job ID strings
        ===============     ====================================================================

        :return: success object
        """
        try:
            url = '{base}/jobs/manage?token={token}'.format(base=self._url, token=self._gis._con.token)
            return Job.manage_jobs(self._gis, url, job_ids, 'Close')
        except:
            self._handle_error(sys.exc_info())

    def create(self, template, count=1, name=None, start=None, end=None, priority=None, description=None,
                   owner=None, group=None, assigned=None, complete=None, notes=None, parent=None):
        """
        Adds a job to the Workflow Manager instance given a user-defined template

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        template            Workflow Manager Job Template ID (required)
        ---------------     --------------------------------------------------------------------
        count               Number of jobs to create
        ---------------     --------------------------------------------------------------------
        name                Job Name
        ---------------     --------------------------------------------------------------------
        start               Job Start Date
        ---------------     --------------------------------------------------------------------
        end                 Job End Date
        ---------------     --------------------------------------------------------------------
        priority            Job Priority Level
        ---------------     --------------------------------------------------------------------
        description         Job Description
        ---------------     --------------------------------------------------------------------
        owner               Job Owner
        ---------------     --------------------------------------------------------------------
        group               Job Group
        ---------------     --------------------------------------------------------------------
        assigned            Initial Job Assignee
        ---------------     --------------------------------------------------------------------
        complete            Integer Percentage Complete
        ---------------     --------------------------------------------------------------------
        notes               Job Notes
        ---------------     --------------------------------------------------------------------
        parent              Parent Job
        ===============     ====================================================================

        :return: Workflow Manager Job Object
        """
        job_object = {
            "numberOfJobs": count,
            "jobName": name,
            "startDate": start,
            "dueDate": end,
            "priority": priority,
            "description": description,
            "ownedBy": owner,
            "assignedType": group,
            "assignedTo": assigned,
            "percentComplete": complete,
            "notes": notes,
            "parentJob": parent
        }
        filtered_object = {}
        for key in job_object:
            if job_object[key] is not None:
                filtered_object[key] = job_object[key]
        url = '{base}/jobTemplates/{template}/job?token={token}'.format(base=self._url, template=template,
                                                                        token=self._gis._con.token)
        post_job = Job(filtered_object, self._gis, url)
        return_obj = post_job.post()
        return return_obj['jobIds']

    def delete_attachment(self, job_id, attachment_id):
        """
        Deletes a job attachment given a job ID and attachment ID

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        job_id              Job ID
        ---------------     --------------------------------------------------------------------
        attachment_id       Attachment ID
        ===============     ====================================================================

        :return: status code
        """
        try:
            res = Job.delete_attachment(self._gis,
                                         '{base}/jobs/{jobId}/attachments/{attachmentId}?token={token}'.format(
                                             base=self._url, jobId=job_id, attachmentId=attachment_id,
                                             item=self._item.id, token=self._gis._con.token))
            return res
        except:
            self._handle_error(sys.exc_info())

    def diagram(self, id):
        """
        Returns the job diagram for the user-defined job

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        id                  Job ID
        ===============     ====================================================================

        :return: Workflow Manager Job Diagram Object
        """
        try:
            return JobDiagram.get(self._gis, '{base}/jobs/{job}/diagram'.format(base=self._url, job=id),
                                  {"token": self._gis._con.token})
        except:
            self._handle_error(sys.exc_info())

    def get(self, id):
        """
        Returns an active job with the given ID

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        id                  Job ID
        ===============     ====================================================================

        :return: Workflow Manager Job Object
        """    
        try:
            url = f'{self._url}/jobs/{id}'
            job_dict = ast.literal_eval(str(self._gis._con.get(url, {"token": self._gis._con.token})).encode('cp850', 'replace').decode('utf-8'))
            return Job(job_dict, self._gis, self._url)
        except:
            self._handle_error(sys.exc_info())

    def search(self, query=None, search_string=None, fields=None, display_names=[], sort_by=[], num=10,
                    start_num=0):
        """
        Runs a search against the jobs stored inside the Workflow Manager instance

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        query               SQL query to search against (e.g. "priority='High'")
        ---------------     --------------------------------------------------------------------
        search_str          Search string to search against (e.g. "High")
        ---------------     --------------------------------------------------------------------
        fields              Field list to return
        ---------------     --------------------------------------------------------------------
        display_names       Display names for the return fields
        ---------------     --------------------------------------------------------------------
        sort_by             Field to sort by (e.g. {'field': 'priority', 'sortOrder': 'Asc'})
        ---------------     --------------------------------------------------------------------
        num                 Number of return results
        ---------------     --------------------------------------------------------------------
        start_num           Index of first return value
        ===============     ====================================================================

        :return: List of search results
        """
        try:
            search_object = {
                'q': query,
                'search': search_string,
                'num': num,
                'displayNames': display_names,
                'start': start_num,
                'sortFields': sort_by,
                'fields': fields
            }
            url = '{base}/jobs/search?token={token}'.format(base=self._url, token=self._gis._con.token)
            return Job.search(self._gis, url, search_object)
        except:
            self._handle_error(sys.exc_info())

    def update(self, job_id, update_object):
        """
        Updates a job object by ID

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        job_id              ID for the job to update
        ---------------     --------------------------------------------------------------------
        update_object       an object containing the fields and new values to add to the job
        ===============     ====================================================================

        :return: success object
        """
        try:
            current_job = self.get(job_id).__dict__
            for k in update_object.keys():
                current_job[k] = update_object[k]
            url = '{base}/jobs/{jobId}/update?token={token}'.format(base=self._url, jobId=job_id,
                                                                    token=self._gis._con.token)
            new_job = Job(current_job, self._gis, url)
            return new_job.post()
        except:
            self._handle_error(sys.exc_info())

    def upgrade(self, job_ids):
        """
        Upgrades a single or multiple jobs with specific JobIDs

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        job_ids             A list of job ID strings
        ===============     ====================================================================

        :return: success object
        """
        try:
            url = '{base}/jobs/manage?token={token}'.format(base=self._url, token=self._gis._con.token)
            return Job.manage_jobs(self._gis, url, job_ids, 'Upgrade')
        except:
            self._handle_error(sys.exc_info())


class WorkflowManager:
    """
    Represents a connection to a Workflow Manager instance or item.

    Users create, update, delete workflow diagrams, job templates and jobs
    or the various other properties with a workflow item.

    ===============     ====================================================================
    **Argument**        **Description**
    ---------------     --------------------------------------------------------------------
    item                The Workflow Manager Item
    ===============     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE: Creating a WorkflowManager object from a workflow item

        from arcgis.workflow import WorkflowManager
        from arcgis.gis import GIS

        # connect to your GIS and get the web map item
        gis = GIS(url, username, password)
        wf_item = gis.content.get('1234abcd_workflow item id')

        # create a WorkflowManager object from the workflow item
        wm = WorkflowManager(wf_item)
        type(wm)
        >> arcgis.workflowmanager.WorkflowManager

        # explore the users in this workflow using the 'users' property
        wm.users
        >> [{}...{}]  # returns a list of dictionaries representing each user
    """

    def __init__(self, item):
        """initializer"""
        if item is None:
            raise ValueError("Item cannot be None")
        self._item = item
        self._gis = item._gis
        if self._gis.users.me is None:
            raise ValueError("An authenticated `GIS` is required.")

        self.job_manager = JobManager(item)

        self._url = self._wmx_server_url[0]
        if self._url is None:
            raise ValueError("No WorkflowManager Registered with your Organization")
        if not any(prov.itemid == '50a5f00bcc574358b15eab0e2bdadf39' for prov in self._gis.users.me.provisions):
            raise ValueError("No Workflow Manager license is available for the current user")

    def _handle_error(self, info):
        """Basic error handler - separated into a function to allow for expansion in future releases"""
        error_class = info[0]
        error_text = info[1]
        raise Exception(error_text)

    @property
    def _wmx_server_url(self):
        """locates the WMX server"""
        baseurl = self._gis._portal.resturl
        res = self._gis._con.get(f"{baseurl}/portals/self/servers", {'f': 'json'})
        for s in res['servers']:
            server_functions = [x.strip() for x in s.get("serverFunction", "").lower().split(",")]
            if 'workflowmanager' in server_functions:
                self._url = s.get("url", None)
                self._private_url = s.get("adminUrl", None)
                if self._url is None:
                    raise RuntimeError("Cannot find a WorkflowManager Server")
                self._url += f"/workflow/{self._item.id}"
                self._private_url += f"/workflow/{self._item.id}"
                return self._url, self._private_url
        return None

    @property
    def jobs(self):
        """
        The job manager for a workflow item. See :class:`~arcgis.workflowmanager.JobManager`.
        """

        return self.job_manager

    def evaluate_arcade(self, expression, context = None, context_type = "BaseContext", mode = "Standard"):
        """
        Evaluates an arcade expression
        
        ======================  ===============================================================
        **Argument**            **Description**
        ----------------------  ---------------------------------------------------------------
        expression              Required String.
        ----------------------  ---------------------------------------------------------------
        context                 Optional String.
        ----------------------  ---------------------------------------------------------------
        context_type            Optional String.
        ----------------------  ---------------------------------------------------------------
        mode                    Optional String. 
        ======================  ===============================================================
        
        :returns: String
        """
        url = f"{self._url}/evaluateArcade?token={self._gis._con.token}"
        params = {
            "expression": expression,
            "contextType": context_type,
            "context": context,
            "parseMode": mode
        }
        res = self._gis._con.post(url, params=params, json_encode=False, post_json=True)
        return res.get("result", None)

    @property
    def wm_roles(self):
        """
        Returns an array of user roles available in the local Workflow Manager instance

        :return: array
        """
        try:
            role_array = ast.literal_eval(str(self._gis._con.get('{base}/community/roles'.format(base=self._url),
                                                           params={"token": self._gis._con.token})['roles']).encode(
                'cp850', 'replace').decode('utf-8'))
            return_array = [WMRole(r) for r in role_array]
            return return_array
        except:
            self._handle_error(sys.exc_info())

    @property
    def users(self):
        """
        Returns an array of all user profiles stored in Workflow Manager

        :return: array
        """
        try:
            user_array = ast.literal_eval(str(self._gis._con.get('{base}/community/users'.format(base=self._url),
                                                           params={"token": self._gis._con.token})['users']).encode(
                'cp850', 'replace').decode('utf-8'))
            return_array = [self.user(u['username']) for u in user_array]
            return return_array
        except:
            self._handle_error(sys.exc_info())

    @property
    def settings(self):
        """
        Returns an array of all settings for the Workflow Manager instance

        :return: array
        """
        try:
            return ast.literal_eval(str(
                self._gis._con.get('{base}/settings'.format(base=self._url), params={"token": self._gis._con.token})[
                    'settings']).encode('cp850', 'replace').decode('utf-8'))
        except:
            self._handle_error(sys.exc_info())

    @property
    def assignable_users(self):
        """
        Get all assignable users for a user in the workflow system

        :return: A list of the assignable user objects
        """
        try:
            user_array = ast.literal_eval(str(self._gis._con.get('{base}/community/usersAssign'.format(base=self._url),
                                                           params={"token": self._gis._con.token})['users']).encode(
                'cp850', 'replace').decode('utf-8'))
            return_array = [self.user(u['username']) for u in user_array]
            return return_array
        except:
            self._handle_error(sys.exc_info())

    @property
    def groups(self):
        """
        Returns an array of all user groups stored in Workflow Manager

        :return: array
        """
        try:
            group_array = ast.literal_eval(str(self._gis._con.get('{base}/community/groups'.format(base=self._url),
                                                           params={"token": self._gis._con.token})['groups']).encode(
                'cp850', 'replace').decode('utf-8'))
            return_array = [self.group(g['id']) for g in group_array]
            return return_array
        except:
            self._handle_error(sys.exc_info())

    @property
    def assignable_groups(self):
        """
        Get portal groups associated with Workflow Manager roles, to which the current user
        can assign work based on their Workflow Manager assignment privileges.

        :return: A list of the assignable group objects
        """
        try:
            group_array = ast.literal_eval(str(self._gis._con.get('{base}/community/groupsAssign'.format(base=self._url),
                                                           params={"token": self._gis._con.token})['groups']).encode(
                'cp850', 'replace').decode('utf-8'))
            return_array = [self.group(g['id']) for g in group_array]
            return return_array
        except:
            self._handle_error(sys.exc_info())

    @property
    def searches(self):
        """
        Returns an array of all saved searches.

        :return: List
        """
        try:
            return ast.literal_eval(str(
                self._gis._con.get('{base}/searches'.format(base=self._url), params={"token": self._gis._con.token})[
                    'searches']).encode('cp850', 'replace').decode('utf-8'))
        except:
            self._handle_error(sys.exc_info())

    @property
    def job_templates(self):
        """
        Gets all the job templates in a workflow item.

        :return: List of all current job templates in the Workflow Manager (required information for create_job call)
        """
        try:
            template_array = ast.literal_eval(str(self._gis._con.get('{base}/jobTemplates'.format(base=self._url),
                                                            params={"token": self._gis._con.token})[
                                            'jobTemplates']).encode('cp850', 'replace').decode('utf-8'))
            return_array = [JobTemplate(t, self._gis, self._url) for t in template_array]
            return return_array
        except:
            self._handle_error(sys.exc_info())

    @property
    def diagrams(self):
        """
        Gets the workflow diagrams within the workflow item.

        :return: List of all current diagrams in the Workflow Manager
        """
        try:
            diagram_array = ast.literal_eval(str(
                self._gis._con.get('{base}/diagrams'.format(base=self._url), params={"token": self._gis._con.token})[
                    'diagrams']).encode('cp850', 'replace').decode('utf-8'))
            return_array = [JobDiagram(d, self._gis, self._url) for d in diagram_array]
            return return_array
        except:
            self._handle_error(sys.exc_info())

    def update_settings(self, props):
        """
        Returns an active job with the given ID

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        props               a list of Props objects to update 
                            (Prop object example: {'propName': 'string', 'value': 'string'})
        ===============     ====================================================================

        :return: success object
        """
        url = '{base}/settings?token={token}'.format(base=self._url, token=self._gis._con.token)
        params = {
            'settings': props
        }
        return_obj = json.loads(
            self._gis._con.post(url, params, add_token=False, post_json=True, try_json=False, json_encode=False))
        if 'error' in return_obj:
            self._gis._con._handle_json_error(return_obj['error'], 0)
        elif 'success' in return_obj:
            return return_obj['success']
        return return_obj

    def wm_role(self, name):
        """
        Returns an active role with the given name

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        name                Role Name
        ===============     ====================================================================

        :return: Workflow Manager Role Object
        """
        try:
            return WMRole.get(self._gis,
                            '{base}/community/roles/{role}'.format(base=self._url, role=urllib.parse.quote(name),
                                                                   item=self._item.id), {"token": self._gis._con.token})
        except:
            self._handle_error(sys.exc_info())

    def job_template(self, id):
        """
        Returns a job template with the given ID

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        id                  Job Template ID
        ===============     ====================================================================

        :return: Workflow Manager Job Template Object
        """
        try:
            print(id)
            return JobTemplate.get(self._gis, '{base}/jobTemplates/{jobTemplate}'.format(base=self._url, jobTemplate=id),
                                   {"token": self._gis._con.token})
        except:
            self._handle_error(sys.exc_info())

    def delete_job_template(self, id):
        """
        Deletes a job template with the given ID

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        id                  Job Template ID
        ===============     ====================================================================

        :return: status code
        """
        try:
            res = JobTemplate.delete(self._gis,
                                      '{base}/jobTemplates/{jobTemplate}?token={token}'.format(base=self._url,
                                                                                               jobTemplate=id,
                                                                                               item=self._item.id,
                                                                                               token=self._gis._con.token))
            return res
        except:
            self._handle_error(sys.exc_info())

    def user(self, username):
        """
        Returns a user profile with the given username

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        username            Workflow Manager Username
        ===============     ====================================================================

        :return: Workflow Manager User Object
        """
        try:
            return arcgis.gis.User(self._gis, username)
        except:
            self._handle_error(sys.exc_info())

    def group(self, group_id):
        """
        Returns group information with the given group ID

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        group_id            Workflow Manager Group ID
        ===============     ====================================================================

        :return: Workflow Manager Group Object
        """
        try:
            wmx_group = Group.get(self._gis, '{base}/community/groups/{groupid}'.format(base=self._url, groupid=group_id,
                                                                                   item=self._item.id),
                             {"token": self._gis._con.token})
            arcgis_group = arcgis.gis.Group(self._gis, group_id)
            arcgis_group.roles = wmx_group.roles
            return arcgis_group
        except:
            self._handle_error(sys.exc_info())

    def diagram(self, id):
        """
        Returns the diagram with the given ID

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        id                  Diagram ID
        ===============     ====================================================================

        :return: Workflow Manager Job Diagram Object
        """
        try:
            return JobDiagram.get(self._gis, '{base}/diagrams/{diagram}'.format(base=self._url, diagram=id),
                                  {"token": self._gis._con.token})
        except:
            self._handle_error(sys.exc_info())

    def create_wm_role(self, name, description='', privileges=[]):
        """
        Adds a role to the Workflow Manager instance given a user-defined name

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        name                Role Name (required)
        ---------------     --------------------------------------------------------------------
        description         Role Description
        ---------------     --------------------------------------------------------------------
        privileges          List of privileges associated with the role
        ===============     ====================================================================

        :return: Workflow Manager Role Object
        """
        try:
            url = '{base}/community/roles/{name}?token={token}'.format(base=self._url, name=name,
                                                                       token=self._gis._con.token)
            post_role = WMRole({
                "roleName": name,
                "description": description,
                "privileges": privileges
            })
            return post_role.post(self._gis, url)
        except:
            self._handle_error(sys.exc_info())

    def create_job_template(self, name, priority, id="", category="", job_duration=0,
                            assigned_to="", default_due_date=None, default_start_date=None,
                            start_date_type="CreationDate", diagram_id="", diagram_name="",
                            assigned_type="Unassigned", description="", default_description="",
                            state="Draft", last_updated_by="", last_updated_date=None,
                            extended_property_table_definitions=[]):
        """
        Adds a job template to the Workflow Manager instance given a user-defined name and default priority level

        ====================================     ====================================================================
        **Argument**                             **Description**
        ------------------------------------     --------------------------------------------------------------------
        name                                     Job Template Name (required)
        ------------------------------------     --------------------------------------------------------------------
        priority                                 Default Job Template Priority Level (required)
        ------------------------------------     --------------------------------------------------------------------
        id                                       Job Template ID
        ------------------------------------     --------------------------------------------------------------------
        category                                 Job Template Category
        ------------------------------------     --------------------------------------------------------------------
        job_duration                             Default Job Template Duration
        ------------------------------------     --------------------------------------------------------------------
        assigned_to                              Job Owner
        ------------------------------------     --------------------------------------------------------------------
        default_due_date                         Due Date for Job Template
        ------------------------------------     --------------------------------------------------------------------
        default_start_date                       Start Date for Job Template
        ------------------------------------     --------------------------------------------------------------------
        start_date_type                          Type of Start Date (e.g. creationDate)
        ------------------------------------     --------------------------------------------------------------------
        diagram_id                               Job Template Diagram ID
        ------------------------------------     --------------------------------------------------------------------
        diagram_name                             Job Template Diagram Name
        ------------------------------------     --------------------------------------------------------------------
        assigned_type                            Type of Job Template Assignment
        ------------------------------------     --------------------------------------------------------------------
        description                              Job Template Description
        ------------------------------------     --------------------------------------------------------------------
        default_description                      Default Job Template Description
        ------------------------------------     --------------------------------------------------------------------
        state                                    Default Job Template State
        ------------------------------------     --------------------------------------------------------------------
        last_updated_by                          User Who Last Updated Job Template
        ------------------------------------     --------------------------------------------------------------------
        last_updated_date                        Date of Last Job Template Update
        ------------------------------------     --------------------------------------------------------------------
        extended_property_table_definitions      List of Extended Properties for Job Template
        ====================================     ====================================================================

        :return: Workflow Manager Job Template ID
        """
        try:
            if default_due_date is None:
                default_due_date = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")
            if default_start_date is None:
                default_start_date = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")
            if last_updated_date is None:
                last_updated_date = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")
            url = '{base}/jobTemplates?token={token}'.format(base=self._url, token=self._gis._con.token)
            post_job_template = JobTemplate({
                "jobTemplateId": id,
                "jobTemplateName": name,
                "category": category,
                "defaultJobDuration": job_duration,
                "defaultAssignedTo": assigned_to,
                "defaultDueDate": default_due_date,
                "defaultStartDate": default_start_date,
                "jobStartDateType": start_date_type,
                "diagramId": diagram_id,
                "diagramName": diagram_name,
                "defaultPriorityName": priority,
                "defaultAssignedType": assigned_type,
                "description": description,
                "defaultDescription": default_description,
                "state": state,
                "extendedPropertyTableDefinitions": extended_property_table_definitions,
                "lastUpdatedBy": last_updated_by,
                "lastUpdatedDate": last_updated_date
            })
            return post_job_template.post(self._gis, url)
        except:
            self._handle_error(sys.exc_info())

    def update_job_template(self, template):
        """
        Updates a job template object by ID

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        body                Job Template body - existing Job Template object that inherits required/optional
                            fields.
        ===============     ====================================================================

        :return: success object
        """
        try:
            url = '{base}/jobTemplates/{jobTemplate}?token={token}'.format(base=self._url,
                                                                           jobTemplate=template['job_template_id'],
                                                                           item=self._item.id,
                                                                           token=self._gis._con.token)
            template_object = JobTemplate(template)
            res = template_object.put(self._gis, url)
            return res
        except:
            self._handle_error(sys.exc_info())

    def create_diagram(self, name, steps, display_grid, description="", active=False, annotations=[], data_sources=[]):
        """
        Adds a diagram to the Workflow Manager instance given a user-defined name and array of steps

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        name                Diagram Name (required)
        ---------------     --------------------------------------------------------------------
        steps               List of Step objects associated with the Diagram (required)
        ---------------     --------------------------------------------------------------------
        display_grid        Boolean indicating whether the grid will be displayed in the Diagram (required)
        ---------------     --------------------------------------------------------------------
        description         Diagram description
        ---------------     --------------------------------------------------------------------
        active              Boolean indicating whether the Diagram is active
        ---------------     --------------------------------------------------------------------
        annotations         List of Annotation objects associated with the Diagram
        ---------------     --------------------------------------------------------------------
        data_sources        List of Data Source objects associated with the Diagram
        ===============     ====================================================================

        :return: Workflow Manager Diagram ID
        """
        try:
            url = '{base}/diagrams?token={token}'.format(base=self._url, token=self._gis._con.token)
            post_diagram = JobDiagram({
                "diagramId": "",
                "diagramName": name,
                "description": description,
                "active": active,
                "initialStepId": "",
                "initialStepName": "",
                "steps": steps,
                "dataSources": data_sources,
                "annotations": annotations,
                "displayGrid": display_grid
            })
            return post_diagram.post(self._gis, url)['diagram_id']
        except:
            self._handle_error(sys.exc_info())

    def update_diagram(self, body, delete_draft=True):
        """
        Updates a diagram object by ID

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        body                Diagram body - existing Diagram object that inherits required/optional
                            fields.
        ---------------     --------------------------------------------------------------------
        delete_draft        Boolean - option to delete the Diagram draft (optional)
        ===============     ====================================================================

        :return: success object
        """
        try:
            url = '{base}/diagrams/{diagramid}?token={token}'.format(base=self._url, diagramid=body['diagram_id'],
                                                                     token=self._gis._con.token)
            post_diagram = JobDiagram({
                "diagramId": body['diagram_id'],
                "diagramName": body['diagram_name'],
                "description": (body['description'] if 'description' in body else ''),
                "active": (body['active'] if 'active' in body else False),
                "initialStepId": (body['initial_step_id'] if 'initial_step_id' in body else ''),
                "initialStepName": (body['initial_step_name'] if 'initial_step_name' in body else ''),
                "steps": body['steps'],
                "dataSources": (body['data_sources'] if 'data_sources' in body else []),
                "annotations": (body['annotations'] if 'annotations' in body else ''),
                "displayGrid": body['display_grid']
            })
            res = post_diagram.update(self._gis, url, delete_draft)

            return res
        except:
            self._handle_error(sys.exc_info())

    def delete_diagram(self, id):
        """
        Deletes a diagram object by ID

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        id                  Diagram id (required)
        ===============     ====================================================================

        :return: Workflow Manager Diagram ID
        """
        try:
            url = '{base}/diagrams/{diagramid}?token={token}'.format(base=self._url, diagramid=id,
                                                                     token=self._gis._con.token)
            return JobDiagram.delete(self._gis, url)
        except:
            self._handle_error(sys.exc_info())


class Job(object):
    """
    Helper class for managing Workflow Manager jobs in a workflow item. This class is not created
    by users directly. An instance of this class, can be created by calling `jobs.get(id)` with the appropriate ID
    for the job. `jobs` is available as property on `arcgis.WorkflowManager`.

    """
    _camelCase_to_underscore = _camelCase_to_underscore
    _underscore_to_camelcase = _underscore_to_camelcase

    def __init__(self, init_data, gis=None, url=None):
        self.job_status = self.notes = self.attachments = self.diagram_id = self.end_date = self.due_date = self.description = self.started_date = self.current_steps = self.job_template_name = self.job_template_id = self.extended_properties = self.diagram_name = self.parent_job = self.job_name = self.diagram_version = self.active_versions = self.percent_complete = self.priority = self.job_id = self.created_date = self.created_by = self.closed = self.owned_by = self.start_date = None
        for key in init_data:
            setattr(self, _camelCase_to_underscore(key), init_data[key])
        self._gis = gis
        self._url = url

    def post(self):
        post_dict = {_underscore_to_camelcase(k): v for k, v in self.__dict__.items() if v is not None and not k.startswith('_')}
        return_obj = json.loads(
            self._gis._con.post(self._url, post_dict, add_token=False, post_json=True, try_json=False, json_encode=False))
        if 'error' in return_obj:
            self._gis._con._handle_json_error(return_obj['error'], 0)
        elif 'success' in return_obj:
            return return_obj['success']
        return return_obj

    def search(gis, url, search_object):
        return_obj = json.loads(
            gis._con.post(url, search_object, add_token=False, post_json=True, try_json=False, json_encode=False))
        if 'error' in return_obj:
            gis._con._handle_json_error(return_obj['error'], 0)
        elif 'success' in return_obj:
            return return_obj['success']
        return_obj = { _camelCase_to_underscore(k): v for k, v in return_obj.items() if v is not None and not k.startswith('_')}
        return return_obj

    def get_attachment(self, attachment_id):
        """
        Returns a job attachment given an attachment ID

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        attachment_id       Attachment ID
        ===============     ====================================================================

        :return: Job Attachment
        """

        url = '{base}/jobs/{jobId}/attachments/{attachmentId}'.format(base=self._url, jobId=self.job_id,
                                                            attachmentId=attachment_id)
        return_obj = self._gis._con.get(url, {"token": self._gis._con.token}, try_json=False)
        if 'error' in return_obj:
            self._gis._con._handle_json_error(return_obj['error'], 0)
        elif 'success' in return_obj:
            return return_obj['success']
        return return_obj

    def add_attachment(self, attachment):
        """
        Adds an attachment to the job

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        attachment          Filepath to attachment
        ===============     ====================================================================

        :return: Job Attachment
        """
        url = '{base}/jobs/{jobId}/attachments'.format(base=self._url, jobId=self.job_id)
        return_obj = json.loads(
            self._gis._con.post(url, files={'attachment': attachment}, add_token=True, try_json=False, json_encode=False))
        if 'error' in return_obj:
            self._gis._con._handle_json_error(return_obj['error'], 0)
        return {
            'id': return_obj['url'].split('/')[-1],
            'alias': return_obj['alias']
        }

    def update_attachment(self, attachment_id, alias):
        """
        Updates an attachment alias given a Job ID and attachment ID

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        attachment_id       Attachment ID
        ---------------     --------------------------------------------------------------------
        alias               Alias
        ===============     ====================================================================

        :return: success
        """
        url = '{base}/jobs/{jobId}/attachments/{attachmentid}'.format(base=self._url, jobId=self.job_id,
                                                                          attachmentid=attachment_id)
        post_object = {
            'alias': alias
        }
        return_obj = json.loads(self._gis._con.post(url, params=post_object, try_json=False, json_encode=False))
        if 'error' in return_obj:
            self._gis._con._handle_json_error(return_obj['error'], 0)
        elif 'success' in return_obj:
            return return_obj['success']
        return_obj = { _camelCase_to_underscore(k): v for k, v in return_obj.items() if v is not None and not k.startswith('_')}
        return return_obj

    def delete_attachment(gis, url):
        return_obj = json.loads(gis._con.delete(url, add_token=False, try_json=False))
        if 'error' in return_obj:
            gis._con._handle_json_error(return_obj['error'], 0)
        elif 'success' in return_obj:
            return return_obj['success']
        return_obj = { _camelCase_to_underscore(k): v for k, v in return_obj.items() if v is not None and not k.startswith('_')}
        return return_obj

    def update_step(self, step_id, assigned_type, assigned_to):
        """
        Update the assignment of the current step in a job based on the current user's Workflow Manager assignment privileges

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        step_id             Active Step ID
        ---------------     --------------------------------------------------------------------
        assigned_type       Type of assignment designated (NOTE: Must be User, Group or Unassigned)
        ---------------     --------------------------------------------------------------------
        assigned_to         User to which the active step is assigned
        ===============     ====================================================================

        :return: success object
        """

        if step_id is None:
            step_id = self.currentSteps[0]['step_id']
        url = '{base}/jobs/{jobId}/{stepId}?token={token}'.format(base=self._url, jobId=self.job_id, stepId=step_id,
                                                                      token=self._gis._con.token)
        post_object = {
            'assignedType': assigned_type,
            'assignedTo': assigned_to
        }
        return_obj = json.loads(
            self._gis._con.post(url, params=post_object, post_json=True, try_json=False, json_encode=False))
        if 'error' in return_obj:
            self._gis._con._handle_json_error(return_obj['error'], 0)
        elif 'success' in return_obj:
            return return_obj['success']
        return_obj = { _camelCase_to_underscore(k): v for k, v in return_obj.items() if v is not None and not k.startswith('_')}
        return return_obj

    def set_current_step(self, step_id):
        """
        Sets a single step to be the active step on the job. The ability to set a step as current is controlled by the **workflowSetStepCurrent** privilege.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        step_id             Active Step ID
        ===============     ====================================================================

        :return: success object
        """

        url = '{base}/jobs/{jobId}/action?token={token}'.format(base=self._url, jobId=self.job_id,
                                                                    token=self._gis._con.token)
        post_object = {
            'type': 'SetCurrentStep',
            'stepIds': [
                step_id
            ]
        }
        return_obj = json.loads(
            self._gis._con.post(url, params=post_object, post_json=True, try_json=False, json_encode=False))
        if 'error' in return_obj:
            self._gis._con._handle_json_error(return_obj['error'], 0)
        elif 'success' in return_obj:
            return return_obj['success']
        return_obj = { _camelCase_to_underscore(k): v for k, v in return_obj.items() if v is not None and not k.startswith('_')}
        return return_obj

    @property
    def history(self):
        """
        Gets the history of a job given job ID

        :return: success object
        """

        url = '{base}/jobs/{jobId}/history?token={token}'.format(base=self._url, jobId=self.job_id,
                                                                    token=self._gis._con.token)
        return_obj = ast.literal_eval(str(self._gis._con.get(url)).encode('cp850', 'replace').decode('utf-8'))
        if 'success' in return_obj:
            return return_obj['success']
        return_obj = { _camelCase_to_underscore(k): v for k, v in return_obj.items() if v is not None and not k.startswith('_')}
        return return_obj

    @property
    def location(self):
        """
        Returns the job location for the user-defined job

        :return: Workflow Manager Job Location Object
        """

        return JobLocation.get(self._gis, '{base}/jobs/{job}/location'.format(base=self._url, job=self.job_id),
                                {"token": self._gis._con.token})

    def manage_jobs(gis, url, ids, action):
        post_object = {
            'jobIds': ids,
            'type': action
        }
        return_obj = json.loads(
            gis._con.post(url, params=post_object, post_json=True, try_json=False, json_encode=False))
        if 'error' in return_obj:
            gis._con._handle_json_error(return_obj['error'], 0)
        elif 'success' in return_obj:
            return return_obj['success']
        return_obj = { _camelCase_to_underscore(k): v for k, v in return_obj.items() if v is not None and not k.startswith('_')}
        return return_obj


class WMRole(object):
    """
    Represents a Workflow Manager Role object with accompanying GET, POST, and DELETE methods
    ===============     ====================================================================
    **Argument**        **Description**
    ---------------     --------------------------------------------------------------------
    init_data           data object representing relevant parameters for GET or POST calls
    ===============     ====================================================================
    """
    _camelCase_to_underscore = _camelCase_to_underscore
    _underscore_to_camelcase = _underscore_to_camelcase

    def __init__(self, init_data):
        self.privileges = self.roleName = self.description = None
        for key in init_data:
            setattr(self, _camelCase_to_underscore(key), init_data[key])

    def get(gis, url, params):
        role_dict = ast.literal_eval(str(gis._con.get(url, params)).encode('cp850', 'replace').decode('utf-8'))
        return WMRole(role_dict)

    def post(self, gis, url):
        post_dict = {_underscore_to_camelcase(k): v for k, v in self.__dict__.items() if v is not None}
        return_obj = json.loads(
            gis._con.post(url, post_dict, add_token=False, post_json=True, try_json=False, json_encode=False))
        if 'error' in return_obj:
            gis._con._handle_json_error(return_obj['error'], 0)
        elif 'success' in return_obj:
            return return_obj['success']
        return return_obj


class JobTemplate(object):
    """
    Represents a Workflow Manager Job Template object with accompanying GET, POST, and DELETE methods
    ===============     ====================================================================
    **Argument**        **Description**
    ---------------     --------------------------------------------------------------------
    init_data           data object representing relevant parameters for GET or POST calls
    ===============     ====================================================================
    """
    _camelCase_to_underscore = _camelCase_to_underscore
    _underscore_to_camelcase = _underscore_to_camelcase

    def __init__(self, init_data, gis=None, url=None):
        for key in init_data:
            setattr(self, _camelCase_to_underscore(key), init_data[key])
        self._gis = gis
        self._url = url
    
    def __getattr__(self, item):
        possible_fields = ['default_assigned_to', 'last_updated_by', 'diagram_id', 'extended_property_table_definitions', 'description', 'job_template_name', 'job_template_id', 'default_start_date', 'default_priority_name', 'last_updated_date', 'job_start_date_type', 'diagram_name', 'default_job_duration', 'default_due_date', 'state', 'category', 'default_assigned_type', 'default_description']
        gis = object.__getattribute__(self, '_gis')
        url = object.__getattribute__(self, '_url')
        id = object.__getattribute__(self, 'job_template_id')
        full_object = ast.literal_eval(str(gis._con.get(url, {"token": gis._con.token})).encode('cp850', 'replace').decode('utf-8'))
        try:
            setattr(self, _camelCase_to_underscore(item), full_object[item])
            return full_object[item]
        except KeyError:
            if item in possible_fields:
                setattr(self, _camelCase_to_underscore(item), None)
                return None
            else:
                raise KeyError(f'The attribute "{item}" is invalid for Job Templates')

    def get(gis, url, params):
        job_template_dict = ast.literal_eval(str(gis._con.get(url, params)).encode('cp850', 'replace').decode('utf-8'))
        return JobTemplate(job_template_dict, gis, url)

    def put(self, gis, url):
        put_dict = {_underscore_to_camelcase(k): v for k, v in self.__dict__.items() if v is not None}
        return_obj = json.loads(gis._con.put(url, put_dict, add_token=False, post_json=True, try_json=False,
                                  json_encode=False))
        if 'error' in return_obj:
            gis._con._handle_json_error(return_obj['error'], 0)
        elif 'success' in return_obj:
            return return_obj['success']
        return_obj = { _camelCase_to_underscore(k): v for k, v in return_obj.items() if v is not None and not k.startswith('_')}
        return return_obj

    def post(self, gis, url):
        post_dict = {_underscore_to_camelcase(k): v for k, v in self.__dict__.items() if v is not None}
        return_obj = json.loads(
            gis._con.post(url, post_dict, add_token=False, post_json=True, try_json=False, json_encode=False))
        if 'error' in return_obj:
            gis._con._handle_json_error(return_obj['error'], 0)
        elif 'success' in return_obj:
            return return_obj['success']
        return return_obj['jobTemplateId']

    def delete(gis, url):
        return_obj = json.loads(gis._con.delete(url, add_token=False, try_json=False))
        if 'error' in return_obj:
            gis._con._handle_json_error(return_obj['error'], 0)
        elif 'success' in return_obj:
            return return_obj['success']
        return_obj = { _camelCase_to_underscore(k): v for k, v in return_obj.items() if v is not None and not k.startswith('_')}
        return return_obj


class Group(object):
    """
    Represents a Workflow Manager Group object with accompanying GET, POST, and DELETE methods
    ===============     ====================================================================
    **Argument**        **Description**
    ---------------     --------------------------------------------------------------------
    init_data           data object representing relevant parameters for GET or POST calls
    ===============     ====================================================================
    """
    _camelCase_to_underscore = _camelCase_to_underscore

    def __init__(self, init_data):
        self.roles = None
        for key in init_data:
            setattr(self, _camelCase_to_underscore(key), init_data[key])

    def get(gis, url, params):
        group_dict = ast.literal_eval(str(gis._con.get(url, params)).encode('cp850', 'replace').decode('utf-8'))
        return Group(group_dict)


class JobDiagram(object):
    """
    Helper class for managing Workflow Manager jobs diagrams in a workflow item. This class is not created
    by users directly. An instance of this class, can be created by calling `diagram.get(id)` with the appropriate ID
    for the diagram. `diagrams` is available as property on `arcgis.WorkflowManager`.

    """
    _camelCase_to_underscore = _camelCase_to_underscore
    _underscore_to_camelcase = _underscore_to_camelcase

    def __init__(self, init_data, gis=None, url=None):
        for key in init_data:
            setattr(self, _camelCase_to_underscore(key), init_data[key])
        self._gis = gis
        self._url = url
    
    def __getattr__(self, item):
        possible_fields = ['display_grid', 'diagram_version', 'diagram_name', 'diagram_id', 'description', 'annotations', 'initial_step_id', 'data_sources', 'steps', 'initial_step_name']
        gis = object.__getattribute__(self, '_gis')
        url = object.__getattribute__(self, '_url')
        id = object.__getattribute__(self, 'diagram_id')
        full_object = ast.literal_eval(str(gis._con.get(url, {"token": gis._con.token})).encode('cp850', 'replace').decode('utf-8'))
        try:
            setattr(self, _camelCase_to_underscore(item), full_object[item])
            return full_object[item]
        except KeyError:
            if item in possible_fields:
                setattr(self, _camelCase_to_underscore(item), None)
                return None
            else:
                raise KeyError(f'The attribute "{item}" is invalid for Diagrams')

    def get(gis, url, params):
        job_diagram_dict = ast.literal_eval(str(gis._con.get(url, params)).encode('cp850', 'replace').decode('utf-8'))
        return JobDiagram(job_diagram_dict, gis, url)

    def post(self, gis, url):
        post_dict = {_underscore_to_camelcase(k): v for k, v in self.__dict__.items() if v is not None}
        return_obj = json.loads(
            gis._con.post(url, post_dict, add_token=False, post_json=True, try_json=False, json_encode=False))
        if 'error' in return_obj:
            gis._con._handle_json_error(return_obj['error'], 0)
        elif 'success' in return_obj:
            return return_obj['success']
        return_obj = { _camelCase_to_underscore(k): v for k, v in return_obj.items() if v is not None and not k.startswith('_')}
        return return_obj

    def update(self, gis, url, delete_draft):
        clean_dict = {_underscore_to_camelcase(k): v for k, v in self.__dict__.items() if v is not None}
        post_object = {
            "deleteDraft": delete_draft,
            "diagram": clean_dict
        }
        return_obj = json.loads(
            gis._con.post(url, post_object, add_token=False, post_json=True, try_json=False, json_encode=False))
        if 'error' in return_obj:
            gis._con._handle_json_error(return_obj['error'], 0)
        elif 'success' in return_obj:
            return return_obj['success']
        return_obj = { _camelCase_to_underscore(k): v for k, v in return_obj.items() if v is not None and not k.startswith('_')}
        return return_obj

    def delete(gis, url):
        return_obj = json.loads(gis._con.delete(url, add_token=False, try_json=False))
        if 'error' in return_obj:
            gis._con._handle_json_error(return_obj['error'], 0)
        elif 'success' in return_obj:
            return return_obj['success']
        return_obj = { _camelCase_to_underscore(k): v for k, v in return_obj.items() if v is not None and not k.startswith('_')}
        return return_obj


class JobLocation(object):
    """
    Represents a Workflow Manager Job Location object with accompanying GET, POST, and DELETE methods
    ===============     ====================================================================
    **Argument**        **Description**
    ---------------     --------------------------------------------------------------------
    init_data           data object representing relevant parameters for GET or POST calls
    ===============     ====================================================================
    """
    _camelCase_to_underscore = _camelCase_to_underscore
    def __init__(self, init_data):
        self.geometry = self.geometryType = None
        for key in init_data:
            setattr(self, _camelCase_to_underscore(key), init_data[key])

    def get(gis, url, params):
        job_location_dict = ast.literal_eval(str(gis._con.get(url, params)).encode('cp850', 'replace').decode('utf-8'))
        return JobLocation(job_location_dict)



import uuid
import arcgis
from .._utils._basewidget import _BaseWidget
from .._utils._basewidget import Legend
from .._utils._basewidget import NoDataProperties


class PieChart(_BaseWidget):
    """
    Creates a dashboard Pie Chart widget.

    =========================   ===========================================
    **Argument**                **Description**
    -------------------------   -------------------------------------------
    item                        Required Portal Item object. Item object can
                                be a Feature Layer or a MapWidget.
    -------------------------   -------------------------------------------
    name                        Optional string. Name of the pie chart
                                widget.
    -------------------------   -------------------------------------------
    layer                       Optional integer. Layer number when item is
                                a mapwidget.
    -------------------------   -------------------------------------------
    categories_from             Optional string. Select from groupByValues,
                                features or fields.
    -------------------------   -------------------------------------------
    title                       Optional string. Title or Caption for the
                                widget.
    -------------------------   -------------------------------------------
    description                 Optional string. Description for the widget.
    =========================   ===========================================
    """
    def __init__(self, item, name='PieChart', layer=0, categories_from="groupByValues", title='', description=''):
        super().__init__(name, title, description)

        if item.type not in ['Feature Service', 'mapWidget']:
            raise Exception("Please specify an item")

        self.item = item
        self.type = "pieChartWidget"
        self.layer = layer

        self._labels = False

        self._max_features = None
        self._show_last_update = True
        self._targetid = None

        if categories_from not in ["groupByValues", "features", "fields"]:
            raise Exception("category_from can be groupByValues or features or fields")

        self._data = PieChartData._create_data(categories_from, data_item = item)
        self._pie = PieChartProperties._create_chart()
        self._slices = SliceProperties._slice_properties()
        self._nodata = NoDataProperties._nodata_init()
        self._legend = Legend._create_legend()
        self._outline = OutlineProperties._outline_init()
        self._events = Events._create_events()

    @classmethod
    def _from_json(cls, widget_json):
        gis = arcgis.env.active_gis
        itemid = widget_json["datasets"]["datasource"]["itemid"]
        name = widget_json["name"]
        item = gis.content.get(itemid)
        title = widget_json["caption"]
        description = widget_json["description"]
        pie_chart = PieChart(item, name, title, description)

        return pie_chart

    @property
    def events(self):
        """
        :return: list of events attached to the widget.
        """
        return self._events

    @property
    def data(self):
        """
        :return: Pie Chart Data object. Set data properties, categories and values.
        """
        return self._data
    
    @property
    def legend(self):
        """
        :return: Legend Object, set Visibility and placement
        """
        return self._legend
    
    @property
    def pie(self):
        """
        :return: Pie Object, set various pie properties
        """
        return self._pie
    
    @property
    def outline(self):
        """
        :return: Outline Object, set various outline properties
        """
        return self._outline

    @property
    def slices(self):
        """
        :return: Slices Object, set various slices properties
        """
        return self._slices

    @property
    def no_data(self):
        """
        :return: Nodata Object, set various nodata properties
        """
        return self._nodata

    @property
    def max_features(self):
        """
        :return: max number of features to display.
        """
        return self._max_features

    @max_features.setter
    def max_features(self, value):
        """
        Set max number of features to display.
        """
        self._max_features = int(value)

    @property
    def show_last_update(self):
        """
        :return: show last update or not.
        """
        return self._show_last_update

    @show_last_update.setter
    def show_last_update(self, value):
        """
        Set show last update or not.
        """
        self._show_last_update = bool(value)

    @property
    def labels(self):
        """
        :return: show labels or not.
        """
        return self._labels

    @labels.setter
    def labels(self, value):
        """
        Set show labels or not.
        """
        self._labels = bool(value)

    def _color_picker(self):
        import random
        
        random_number = random.randint(0,16777215)
        hex_number = str(hex(random_number))
        hex_number ='#'+ hex_number[2:]

        return hex_number

    def _convert_to_json(self):

        self._fields_slices = []
        self._statistic_fields = []

        if self.data.categories_from in ["groupByValues", "features"]:
            self.data._get_slices()

        if self._data.categories_from == "fields":
            self._category = None
            self._field_name = "category"
            for category_field in self._data.category_field:
                self._fields_slices.append({"key":str(category_field), "label":str(category_field), "color":self._color_picker()})
                self._statistic_fields.append({"onStatisticField":str(category_field), "outStatisticFieldName":str(category_field), "statisticType":self.data.statistic})
        elif self._data.categories_from == "groupByValues":
            for slices in self.data._slice_fields:
                self._fields_slices.append({"key":str(slices), "label":str(slices), "color":self._color_picker()})
            self._statistic_fields.append({"onStatisticField":str(self.data.statistics_field), "outStatisticFieldName":"value", "statisticType":self.data.statistic})
            self._category = self._data.category_field
        else:
            for slices in self.data._slice_fields:
                self._fields_slices.append({"key":str(slices), "label":str(slices), "color":self._color_picker()})
            self._category = self._data.category_field

        if self.labels == True:
            self._labels_format = "value"
        else:
            self._labels_format = "hide"

        if self.item.type == 'mapWidget':
            wlayer = self.item.layers[self.layer]
            widget_id = self.item._id
            layer_id = wlayer["id"]
            self._datasource = {"id":str(widget_id)+'#'+str(layer_id)}
            self._targetid = self._datasource
        else:
            self._datasource = {
                        "type": "featureServiceDataSource",
                        "itemId": self.item.itemid,
                        "layerId": 0,
                        "table": True
                    }

        json_data = {
            "type": "pieChartWidget",
            "category": {
                "sliceProperties": self._fields_slices,
                "fieldName": self._category if self._category else self._field_name,
                "nullLabel":self._slices.null_label,
                "blankLabel":self._slices.blank_label,
                "defaultColor":self._slices.default_color,
                "nullColor":self._slices.null_color,
                "blankColor":self._slices.blank_color
            },
            "pie": {
                "type":"pie",
                "color":self._pie.text_color,
                "fontSize":self._pie.font_size,
                "titleField":"category",
                "valueField":"absoluteValue",
                "alpha":self._slices.opacity,
                "outlineAlpha":self._outline.opacity,
                "outlineColor":self._outline.color,
                "outlineThickness":self._outline.thickness,
                "innerRadius":self._pie.inner_radius,
                "labelsEnabled":self.labels,
                "labelsFormat":self._labels_format,
                "labelTickAlpha":0.5,
                "labelTickColor":"#fab123",
                "maxLabelWidth":100,
                "startAngle":self._pie.start_angle,
                "autoMargins":False,
                "marginTop":0,
                "marginBottom":0,
                "marginLeft":0,
                "marginRight":0,
                "groupPercent":self._slices.grouping_percent,
                "groupedColor":self._slices.grouping_color
            },
            #"valueField":self._data.value_field,
            "legend":{
                "enabled":self._legend.visibility,
                "format":"value",
                "position":self._legend.placement,
                "markerSize":15,
                "markerType":"circle",
                "align":"center",
                "labelWidth":100,
                "valueWidth":50
            },
            "showBalloon":True,
            "valueFormat":{
                "name":"value",
                "type":"decimal",
                "prefix":True,
                "pattern":"#,###.#"
            },
            "percentageFormat":{
                "name":"percentage",
                "type":"decimal",
                "prefix":False,
                "pattern":"#.##"
            },
            "events":[],
            "selectionMode":"single",
            "categoryType":self._data.categories_from,
            "datasets":[
                    {
                    "type":"serviceDataset",
                    "dataSource":self._datasource,
                    "outFields":["*"],
                    "groupByFields":[self._category] if self._category else [],
                    "orderByFields":[],
                    "statisticDefinitions":self._statistic_fields,
                    "maxFeatures": self.max_features,
                    "querySpatialRelationship":"esriSpatialRelIntersects",
                    "returnGeometry":False,
                    "clientSideStatistics":False,
                    "name":"main"
                }
            ],
            "id": self._id,
            "name": self.name,
            "caption": self.title,
            "description": self.description,
            "showLastUpdate": self.show_last_update,
            "noDataText":self._nodata.text,
            "noDataVerticalAlignment":self._nodata.alignment,
            "showCaptionWhenNoData":self._nodata.show_title,
            "showDescriptionWhenNoData":self._nodata.show_description
        }

        if self.events.enable:
            json_data["events"].append({"type":self.events.type, "actions":self.events.synced_widgets})
            json_data["selectionMode"] = self.events.selection_mode

        if self.background_color:
            json_data["backgroundColor"] = self.background_color,

        if self.text_color:
            json_data["textColor"] = self.text_color

        if self.events.enable:
            json_data["events"].append({"type": self.events.type, "actions": self.events.synced_widgets})
            json_data["selectionMode"] = self.events.selection_mode

        if hasattr(self._data, '_value_field'):
            json_data["valueField"] = self._data._value_field

        return json_data


class PieChartData(object):

    @classmethod
    def _create_data(cls, categories_from, data_item=None):
        piechart_data = PieChartData()

        piechart_data._categories_from = "groupByValues"
        piechart_data._item = data_item
        piechart_data._category_field = None
        piechart_data._objectid_field = data_item.tables[0].properties["objectIdField"]

        if categories_from != "fields":
            piechart_data._value_field = piechart_data._objectid_field
        piechart_data._statistics_field = piechart_data._objectid_field
        piechart_data._parse_dates = True
        piechart_data._statistic = "count"
        piechart_data._filters = []
        
        piechart_data._slice_fields = []

        piechart_data._categories_from = categories_from

        return piechart_data

    @property
    def categories_from(self):
        """
        :return: Categories from groupByValues, features or fields.
        """
        return self._categories_from

    @property
    def parse_dates(self):
        """
        :return: True if input category field of type date is to be parsed.
        """
        return self._parse_dates

    @parse_dates.setter
    def parse_dates(self, value):
        """
        Set True to parse input category fields of type date. For groupByValues and features.
        """
        self._parse_dates = bool(value)
        
    @property
    def objectid_field(self):
        """
        :return: object ID field name.
        """
        return self._objectid_field
    
    @property
    def statistics_field(self):
        """
        :return: Statistics field name.
        """
        return self._statistics_field

    @property
    def category_field(self):
        """
        :return: Category field from dataset. For groupByValues or features.
        """
        return self._category_field

    @category_field.setter
    def category_field(self, value):
        """
        Set category field from dataset. For groupByValues or features pass a single value. For fields pass a list
        """
        self._category_field = value
            
    @property
    def value_field(self):
        """
        :return: Value field from dataset. For features only.
        """
        return self._value_field

    @value_field.setter
    def value_field(self, value):
        """
        Set value field from dataset. For features and groupByValues only.
        """
        if self._categories_from in ["fields"]:
            raise Exception("Can set this attribute for 'features' or 'groupByValues' category only.")

        self._value_field = value

    @property
    def statistic(self):
        """
        :return: statistic used for input category fields. For values from fields.
        """
        return self._statistic

    @statistic.setter
    def statistic(self, value):
        """
        Set statistic to 'count', 'avg', 'min', 'max', 'stddev', 'sum'
        """
        if self._categories_from in ["features"]:
            raise Exception("Can't set this attribute for 'features' category.")

        if value in ['count', 'avg', 'min', 'max', 'stddev', 'sum']:
            self._statistic = value
    
    @property
    def filters(self):
        """
        :return: filters associated with widget
        """
        return self._filters
    
    def add_filter(self, field, join, condition, **kwargs):
        """
        Add filters associated with widget.
        """
        self._filter_field = field
        if join in ["AND", "OR"]:
            self._filter_join = join
        else:
            raise Exception("Please select from 'AND', 'OR'")
        if condition in ["between", "not between", "equal", "not equal", "greater than", "greater than or equal", "less than", "less than or equal", "is null", "is not null"]:
            self._filter_condition = condition
        else:
            raise Exception("Please select the right condition")

        if condition in ["between", "not between"]:
            self._val1 = kwargs.get('start')
            self,_val2 = kwargs.get('end')
            self._filters.append({"filtertype":self._filter_join, "field":self._filter_field, "operator":self._filter_condition, "start":self._val1, "end":self._val2})
        else:
            raise Exception("Please provide 'start' and 'end' values as parameters")

        if condition in ["equal", "not equal", "greater than", "greater than or equal", "less than", "less than or equal"]:
            self._val = kwargs.get('value')
            self._filters.append({"filtertype":self._filter_join, "field":self._filter_field, "operator":self._filter_condition, "value":self._val})
        else:
            raise Exception("Please provide a 'value' parameter for comparison")

    def _get_slices(self):
        statistics = self.statistic
        c_field = self.category_field
        v_field = self.value_field

        x = self._item.tables[0].query(group_by_fields_for_statistics=c_field ,out_statistics=[{"statisticType":statistics,"onStatisticField":v_field,"outStatisticFieldName":"value"}], as_df=True)
        self._slice_fields = x[c_field]

class PieChartProperties(object):

    @classmethod
    def _create_chart(cls, text_color="#000000", font_size=12, start_angle=90, inner_radius=0, hover_text=True):

        chart = PieChartProperties()

        chart._text_color= text_color
        chart._font_size = font_size
        chart._start_angle = start_angle
        chart._inner_radius = inner_radius
        chart._hover_text = hover_text

        chart.text_color = text_color
        chart.font_size = font_size
        chart.start_angle = start_angle
        chart.inner_radius = inner_radius
        chart.hover_text = hover_text

        return chart

    @property
    def text_color(self):
        """
        :return: Text Color of Pie Chart.
        """
        return self._text_color

    @text_color.setter
    def text_color(self, value):
        """
        Set Text Color of Pie Chart.
        """
        self._text_color = str(value)

    @property
    def font_size(self):
        """
        :return: font size of Pie Chart.
        """
        return self._font_size

    @font_size.setter
    def font_size(self, value):
        """
        Set font size value.
        """
        self._font_size = value

    @property
    def start_angle(self):
        """
        :return: Start Angle of Pie Chart.
        """
        return self._start_angle

    @start_angle.setter
    def start_angle(self, value):
        """
        Set Start Angle value.
        """
        self._start_angle = value

    @property
    def inner_radius(self):
        """
        :return: Inner Radius of Pie Chart.
        """
        return self._inner_radius

    @inner_radius.setter
    def inner_radius(self, value):
        """
        Set Inner Radius value.
        """
        self._inner_radius = value

    @property
    def hover_text(self):
        """
        :return: Show text on hovering or not.
        """
        return self._hover_text

    @hover_text.setter
    def hover_text(self, value):
        """
        Set Show text on hovering or not.
        """
        self._hover_text = value

class SliceProperties(object):

    @classmethod
    def _slice_properties(cls, opacity=1.0, default_color="#ffffff", null_color="#ffffff", blank_color="#000000", grouping_color="#000000", grouping_percent=0, null_label='Null', blank_label='Blank'):
        slices = SliceProperties()

        slices._opacity = 1.0
        slices._default_color = "#ffffff"

        slices._null_color = "#ffffff"
        slices._null_label = 'Null'

        slices._blank_color = "#000000"
        slices._blank_label = 'Blank'

        slices._grouping_percent = 0
        slices._grouping_color = "#000000"

        slices.opacity = opacity
        slices.default_color = default_color

        slices.null_color = null_color
        slices.null_label = null_label

        slices.blank_color = blank_color
        slices.blank_label = blank_label

        slices.grouping_percent = grouping_percent
        slices.grouping_color = grouping_color

        return slices

    @property
    def opacity(self):
        """
        :return: slices opacity value.
        """
        return self._opacity
    
    @opacity.setter
    def opacity(self, value):
        """
        Set slices opactiy value.
        """
        self._opacity = value

    @property
    def default_color(self):
        """
        :return: default color of slices.
        """
        return self._default_color

    @default_color.setter
    def default_color(self, value):
        """
        Set default color of slices.
        """
        self._default_color = str(value)

    @property
    def null_color(self):
        """
        :return: color of null slices.
        """
        return self._null_color

    @null_color.setter
    def null_color(self, value):
        """
        Set color of null slices.
        """
        self._null_color = str(value)

    @property
    def blank_color(self):
        """
        :return: color of blank slices.
        """
        return self._blank_color

    @blank_color.setter
    def blank_color(self, value):
        """
        Set color of blank slices.
        """
        self._blank_color = str(value)

    @property
    def grouping_color(self):
        """
        :return: grouping color.
        """
        return self._grouping_color

    @grouping_color.setter
    def grouping_color(self, value):
        """
        Set grouping color.
        """
        self._grouping_color = str(value)

    @property
    def null_label(self):
        """
        :return: label of null slices.
        """
        return self._null_label

    @null_label.setter
    def null_label(self, value):
        """
        Set label of null slices.
        """
        self._null_label = str(value)

    @property
    def blank_label(self):
        """
        :return: label of blank slices.
        """
        return self._blank_label

    @blank_label.setter
    def blank_label(self, value):
        """
        Set label of blank slices.
        """
        self._blank_label = str(value)

    @property
    def grouping_percent(self):
        """
        :return: grouping percentage.
        """
        return self._grouping_percent

    @grouping_percent.setter
    def grouping_percent(self, value):
        """
        Set grouping percentage.
        """
        self._grouping_percent = value

class OutlineProperties(object):

    @classmethod
    def _outline_init(cls, opacity=0.4, thickness=0.2, color="#000000"):
        outline = OutlineProperties()

        outline._opacity = opacity
        outline._thickness = thickness
        outline._color = color

        outline.opacity = opacity
        outline.thickness = thickness
        outline.color = color

        return outline

    @property
    def opacity(self):
        """
        :return: outline opacity.
        """
        return self._opacity

    @opacity.setter
    def opacity(self, value):
        """
        Set outline opacity.
        """
        self._opacity = value

    @property
    def thickness(self):
        """
        :return: outline thickness.
        """
        return self._thickness

    @thickness.setter
    def thickness(self, value):
        """
        Set outline thickness.
        """
        self._thickness = value

    @property
    def color(self):
        """
        :return: outline color.
        """
        return self._color

    @color.setter
    def color(self, value):
        """
        Set outline color.
        """
        self._color = str(value)
 
class Events(object):

    @classmethod
    def _create_events(cls, enable=False):
        events = Events()

        events._enable = False
        events._selection_mode = "single"
        events._type = "selectionChanged"
        events._actions = []

        events.enable = enable

        return events

    @property
    def selection_mode(self):
        """
        :return: Selection mode of events.
        """
        return self._selection_mode

    @selection_mode.setter
    def selection_mode(self, value):
        """
        Set Selection mode of events.
        """
        if value in ["single", "multi"]:
            self._selection_mode = value
        else:
            raise Exception("Please specify selection_mode from 'single' and 'multi'")

    @property
    def enable(self):
        """
        :return: if events are enabled or not.
        """
        return self._enable

    @enable.setter
    def enable(self, value):
        """
        Set if events are enabled or not.
        """
        self._enable = bool(value)

    @property
    def type(self):
        """
        :return: type of trigger for event.
        """
        return self._type

    @property
    def synced_widgets(self):
        """
        :return: List of synced widgets.
        """
        return self._actions

    def sync_map(self, action_type, widget):
        """
        Synchronize a mapWidget with PieChart for triggered events.

        =========================   ===========================================
        **Argument**                **Description**
        -------------------------   -------------------------------------------
        action_type                 Required string. Actions can be one of 
                                    "zoom", "flash", "show_popup", "pan".
        -------------------------   -------------------------------------------
        widget                      Required MapWidget item. Name of the map
                                    widget.
        =========================   ===========================================
        """
        if self.enable == False:
            raise Exception("Please enable events")
        else:
            if widget.type == "mapWidget":
                if action_type in ["zoom", "flash", "show_popup", "pan"]:
                    self._actions.append({"type": action_type, "targetId": widget._id})
                elif action_type == "filter":
                    if self._targetid is not None:
                        self._actions.append({"type":action_type, "by":"whereClause", "targetId":self._targetid})
                    else:
                        raise Exception("This operation is not suitable for given dataSource.")
                else:
                    raise Exception("Please select action_type from 'zoom', 'flash', 'show_popup' and 'pan'")
            else:
                raise Exception("Please select a map widget")

    def sync_widget(self, widgets):
        """
        Synchronize non-mapWidget type widgets with PieChart for triggered events.

        =========================   ===========================================
        **Argument**                **Description**
        -------------------------   -------------------------------------------
        widget                      Required widget item or list of widget items
                                    .Name of the widgets to be synced.
        =========================   ===========================================
        """
        if self.enable == False:
            raise Exception("Please enable events")

        else:
            if isinstance(widgets, list):
                for widget in widgets:
                    if widget.type == "mapWidget":
                        raise Exception("Use sync_map method to add actions for map widgets") ##duplicate or erase
                    else:
                        action_type = "filter"
                        widget_id = str(widget._id)+'#main'
                        self._actions.append({"type":action_type, "by":"whereClause", "targetId":widget_id})
            else:
                if widgets.type == "mapWidget":
                    raise Exception("Use sync_map method to add actions for map widgets") ##duplicate or erase
                else:
                    action_type = "filter"
                    widget_id = str(widgets._id)+'#main'
                    self._actions.append({"type":action_type, "by":"whereClause", "targetId":widget_id})



import os
import random
import json
import pickle
import warnings
import math
from pathlib import Path

import arcgis
from arcgis.features import FeatureLayer

from .._utils.tabular_data import TabularDataObject

HAS_SK_LEARN = True
try:
    import sklearn
    from sklearn import *
except:
    HAS_SK_LEARN = False

HAS_FAST_PROGRESS = True
try:
    from fastprogress.fastprogress import progress_bar
except:
    HAS_FAST_PROGRESS = False

_PROTOCOL_LEVEL = 2


def _get_model_type(model_type):
    if not model_type:
        raise Exception("Invalid model type.")

    if not isinstance(model_type, str):
        return model_type

    if not model_type.startswith('sklearn.'):
        raise Exception("Invalid model_type.")

    model_type = model_type.replace('sklearn.', '')

    module = model_type.split('.')[0]

    if len(model_type.split('.')) > 1:
        model = model_type.split('.')[1]
    else:
        raise Exception("Invalid model_type.")

    if not hasattr(sklearn, module) or \
            not hasattr(getattr(sklearn, module), model):
        raise Exception("Invalid model_type.")

    model = getattr(getattr(sklearn, module), model)

    return model


def raise_data_exception():
    raise Exception("Cannot call this function without data.")


class MLModel(object):
    """
    Creates a machine learning model based on its implementation from scikit-learn.
    For supervised learning:
    Refer https://scikit-learn.org/stable/supervised_learning.html#supervised-learning
    For unsupervised learning:
    1. Clustering Models
    2. Gaussian Mixture Models
    3. Novelty and outlier detection
    Refer https://scikit-learn.org/stable/unsupervised_learning.html

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required TabularDataObject. Returned data object from
                            `prepare_tabulardata` function.
    ---------------------   -------------------------------------------
    model_type              Required string path to the module.
                            For example for SVM:
                                sklearn.svm.SVR or sklearn.svm.SVC
                            For tree:
                                sklearn.tree.DecisionTreeRegressor or sklearn.tree.DecisionTreeClassifier
    ---------------------   -------------------------------------------
    **kwargs                model_type specific arguments.
                            Refer Parameters section
                            https://scikit-learn.org/stable/supervised_learning.html#supervised-learning
    =====================   ===========================================

    :returns: `MLModel` Object
    """

    def __init__(self, data, model_type, **kwargs):
        if not HAS_SK_LEARN:
            raise Exception("This module requires scikit-learn.")

        self._data = data
        self._training_data, self._training_labels, self._validation_data, self._validation_labels = self._data._ml_data
        if kwargs.get('pretrained_model'):
            self._model = kwargs.get('pretrained_model')
        else:
            model = _get_model_type(model_type)

            if model == sklearn.cluster._kmeans.KMeans:
                if not kwargs.get('n_clusters'):
                    kwargs['n_clusters'] = self._get_number_of_clusters(**kwargs)
            elif model == sklearn.mixture._gaussian_mixture.GaussianMixture:
                if not kwargs.get('n_components'):
                    kwargs['n_components'] = self._get_number_of_components(**kwargs)

            self._model = model(**kwargs)

    def fit(self):
        if (not self._data._is_unsupervised and (self._training_data is None or self._training_labels is None))\
                or (self._data._is_unsupervised and self._training_data is None):
            raise_data_exception()

        if self._data._is_unsupervised:
            self._model.fit(self._training_data)
        else:
            self._model.fit(self._training_data, self._training_labels)

    def show_results(self, rows=5):
        """
        Shows sample results for the model.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional number of rows. By default, 5 rows
                                are displayed.
        =====================   ===========================================
        :returns dataframe
        """
        if (not self._data._is_unsupervised and (self._validation_data is None or self._validation_labels is None)) \
                or (self._data._is_unsupervised and self._validation_data is None):
            raise_data_exception()

        min_size = len(self._validation_data)

        if rows < min_size:
            min_size = rows

        # sample_batch = random.sample(self._data._validation_indexes, min_size)
        sample_batch = random.sample(range(len(self._validation_data)), min_size)
        validation_data_batch = self._validation_data.take(sample_batch, axis=0)
        sample_indexes = [self._data._validation_indexes[i] for i in sample_batch]

        output_labels = self._predict(validation_data_batch)
        df = self._data._dataframe.loc[sample_indexes]#.loc[sample_batch]#.reset_index(drop=True).loc[sample_batch].reset_index(drop=True)

        if self._data._dependent_variable:
            df[self._data._dependent_variable + '_results'] = output_labels
        else:
            df['prediction_results'] = output_labels

        return df.sort_index()

    def score(self):
        """
        :returns output from scikit-learn's model.score(), R2 score in case of regression and Accuracy in case of classification.
        For KMeans returns Opposite of the value of X on the K-means objective.
        """
        if (not self._data._is_unsupervised and (self._validation_data is None or self._validation_labels is None)) \
                or (self._data._is_unsupervised and self._validation_data is None):
            raise_data_exception()

        if self._data._is_unsupervised:
            if hasattr(self._model, 'score'):
                return self._model.score(self._training_data)

            raise Exception("Score function not applicable for unsupervised data")

        return self._model.score(self._validation_data, self._validation_labels)

    def decision_function(self):
        """
        :returns output from scikit-learn's model.decision_function()
        """
        if self._training_data is None:
            raise_data_exception()

        if not hasattr(self._model, 'decision_function'):
            raise Exception("Function not implemented for this model.")

        return self._model.decision_function(self._training_data)

    def mahalanobis(self):
        """
        :returns output from scikit-learn's model.mahalanobis()
        """
        if self._training_data is None:
            raise_data_exception()

        if not hasattr(self._model, 'mahalanobis'):
            raise Exception("Function not implemented for this model.")

        return self._model.mahalanobis(self._training_data)

    def kneighbors(self, X=None, n_neighbors=None, return_distance=True):
        """
        :returns output from scikit-learn's model.kneighbors()
        """
        if not hasattr(self._model, 'kneighbors'):
            raise Exception("Function not implemented for this model.")

        kwargs = {}
        if X:
            kwargs['X'] = X
        elif self._training_data is None:
            raise Exception("No data found")
        else:
            kwargs['X'] = self._training_data

        if n_neighbors:
            kwargs['n_neigbors'] = n_neighbors

        kwargs['return_distance'] = return_distance

        return self._model.kneighbors(**kwargs)

    def predict_proba(self):
        """
        :returns output from scikit-learn's model.predict_proba()
        """

        if not hasattr(self._model, 'predict_proba'):
            raise Exception("Function not implemented for this model.")

        if self._training_data is None:
            raise Exception("No data found.")

        return self._model.predict_proba(self._training_data)

    @property
    def feature_importances_(self):
        """
        :returns output from scikit-learn's model.feature_importances_
        """
        if not hasattr(self._model, 'feature_importances_'):
            raise Exception("Property not implemented for this model.")

        return self._model.feature_importances_

    def save(self, name_or_path):
        """
        Saves the model, creates an Esri Model Definition. Uses pickle to save the model.
        Using protocol level 2. Protocol level is backward compatible.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Folder path to save the model.
        =====================   ===========================================
        :returns dataframe
        """

        if (not self._data._is_unsupervised and (self._training_data is None or self._training_labels is None)) \
                or (self._data._is_unsupervised and self._training_data is None):
            raise_data_exception()

        if '\\' in name_or_path or '/' in name_or_path:
            path = name_or_path
        else:
            path = os.path.join(os.getcwd(), name_or_path)

        if not os.path.exists(os.path.dirname(path)):
            raise Exception("Path doesn't exist")

        if not os.path.exists(path):
            os.mkdir(path)

        base_file_name = os.path.basename(path)

        model_file = os.path.join(path, base_file_name + '.pkl')

        with open(model_file, 'wb') as f:
            f.write(pickle.dumps(self._model, protocol=_PROTOCOL_LEVEL))

        MLModel._save_encoders(self._data._encoder_mapping, path, base_file_name)

        if self._data._procs:
            MLModel._save_transforms(self._data._procs, path, base_file_name)

        self._write_emd(path, base_file_name)

        return Path(path)

    def _write_emd(self, path, base_file_name):
        emd_file = os.path.join(path, base_file_name + '.emd')
        emd_params = {}
        emd_params['version'] = str(sklearn.__version__)
        if not self._data._is_unsupervised:
            emd_params['score'] = self.score()
        emd_params['_is_classification'] = "classification" if self._data._is_classification else "regression"
        emd_params['ModelName'] = type(self._model).__name__
        emd_params['ModelFile'] = base_file_name + '.pkl'
        emd_params['ModelParameters'] = self._model.get_params()
        emd_params['categorical_variables'] = self._data._categorical_variables

        if self._data._dependent_variable:
            emd_params['dependent_variable'] = self._data._dependent_variable

        emd_params['continuous_variables'] = self._data._continuous_variables

        with open(emd_file, 'w') as f:
            f.write(json.dumps(emd_params, indent=4))

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a `MLModel` Object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Esri Model Definition
                                file.
        ---------------------   -------------------------------------------
        data                    Required TabularDataObject or None. Returned data
                                object from `prepare_tabulardata` function or None for
                                inferencing.
        =====================   ===========================================

        :returns: `MLModel` Object
        """
        if not HAS_SK_LEARN:
            raise Exception("This module requires scikit-learn.")

        if not os.path.exists(emd_path):
            raise Exception("Invalid data path.")

        with open(emd_path, 'r') as f:
            emd = json.loads(f.read())

        categorical_variables = emd['categorical_variables']
        dependent_variable = emd.get('dependent_variable', None)
        continuous_variables = emd['continuous_variables']
        model_parameters = emd['ModelParameters']

        if emd['version'] != str(sklearn.__version__):
            warnings.warn(f"Sklearn version has changed. Model Trained using version {emd['version']}")

        _is_classification = True
        if emd['_is_classification'] != "classification":
            _is_classification = False

        encoder_mapping = None
        if categorical_variables:
            encoder_path = os.path.join(os.path.dirname(emd_path), os.path.basename(emd_path).split('.')[0] + '_encoders.pkl')
            if os.path.exists(encoder_path):
                with open(encoder_path, 'rb') as f:
                    encoder_mapping = pickle.loads(f.read())

        column_transformer = None
        transforms_path = os.path.join(os.path.dirname(emd_path), os.path.basename(emd_path).split('.')[0] + '_transforms.pkl')
        if os.path.exists(transforms_path):
            with open(transforms_path, 'rb') as f:
                column_transformer = pickle.loads(f.read())

        if data is None:
            data = TabularDataObject._empty(categorical_variables, continuous_variables, dependent_variable, encoder_mapping, column_transformer)
            data._is_classification = _is_classification

        model_file = os.path.join(os.path.dirname(emd_path), emd['ModelFile'])
        with open(model_file, 'rb') as f:
            model = pickle.loads(f.read())

        return cls(data, emd['ModelName'], pretrained_model=model, **model_parameters)

    def _predict(self, data):
        return self._model.predict(data)

    @staticmethod
    def _save_encoders(encoder_mapping, path, base_file_name):
        if not encoder_mapping:
            return

        encoder_file = os.path.join(path, base_file_name + '_encoders.pkl')
        with open(encoder_file, 'wb') as f:
            f.write(pickle.dumps(encoder_mapping, protocol=_PROTOCOL_LEVEL))

    @staticmethod
    def _save_transforms(column_transformer, path, base_file_name):
        if not column_transformer:
            return

        transforms_file = os.path.join(path, base_file_name + '_transforms.pkl')
        with open(transforms_file, 'wb') as f:
            f.write(pickle.dumps(column_transformer, protocol=_PROTOCOL_LEVEL))

    @property
    def _is_kmeans(self):
        return self._model.__class__ == sklearn.cluster._kmeans.KMeans

    def _get_number_of_clusters(self, **kwargs):
        print("Finding optimum number of clusters")

        from sklearn.cluster import KMeans
        from sklearn.metrics import silhouette_score

        range_n_clusters = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20]

        threshold = 5
        count = 0

        max_score = -1
        max_cluster = 2
        scores = []
        for n_clusters in range_n_clusters:
            # Initialize the clusterer with n_clusters value and a random generator
            # seed of 10 for reproducibility.
            clusterer = KMeans(n_clusters=n_clusters, **kwargs)
            cluster_labels = clusterer.fit_predict(self._training_data)

            # The silhouette_score gives the average value for all the samples.
            # This gives a perspective into the density and separation of the formed
            # clusters
            silhouette_avg = silhouette_score(self._training_data, cluster_labels)
            scores.append(silhouette_avg)

            if silhouette_avg > max_score:
                max_score = silhouette_avg
                max_cluster = n_clusters
                count = 0
            else:
                count = count + 1

            if count == threshold:
                break

        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(1, 1)

        ax.plot(
            range_n_clusters[0:len(scores)],
            scores
        )
        ax.set_xlabel("Cluster")
        ax.set_ylabel("Silhouette scores")
        # ax.set_xscale('log')
        # ax.xaxis.set_major_formatter(plt.FormatStrFormatter('%.0e'))
        ax.plot(
            max_cluster,
            max_score,
            markersize=10,
            marker='o',
            color='red'
        )

        plt.show()

        print(f"Selecting n_clusters={max_cluster}")

        return max_cluster

    def _get_number_of_components(self, **kwargs):
        print("Finding optimum number of components")

        from sklearn.mixture import GaussianMixture

        range_n_components = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20]

        max_score = -1
        max_component = 2

        threshold = 5
        count = 0
        scores = []
        for component in range_n_components:
            gm_model = GaussianMixture(n_components=component, **kwargs)

            gm_model.fit(self._training_data)
            score = gm_model.bic(self._training_data)

            scores.append(score)

            if score > max_score:
                max_component = component
                max_score = score
                count = 0
            else:
                count = count + 1

            if count == threshold:
                break

        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(1, 1)

        ax.plot(
            range_n_components[0:len(scores)],
            scores
        )
        ax.set_xlabel("Cluster")
        ax.set_ylabel("BIC score")
        # ax.set_xscale('log')
        # ax.xaxis.set_major_formatter(plt.FormatStrFormatter('%.0e'))
        ax.plot(
            max_component,
            max_score,
            markersize=10,
            marker='o',
            color='red'
        )

        plt.show()

        print(f"Selecting n_components={max_component}")

        return max_component

    def predict(
            self,
            input_features=None,
            explanatory_rasters=None,
            datefield=None,
            distance_features=None,
            output_layer_name="Prediction Layer",
            gis=None,
            prediction_type='features',
            output_raster_path=None,
            match_field_names=None):
        """

        Predict on data from feature layer, dataframe and or raster data.

        =================================   =========================================================================
        **Argument**                        **Description**
        ---------------------------------   -------------------------------------------------------------------------
        input_features                      Optional Feature Layer or spatial dataframe. Required if prediction_type='features'.
                                            Contains features with location and
                                            some or all fields required to infer the dependent variable value.
        ---------------------------------   -------------------------------------------------------------------------
        explanatory_rasters                 Optional list. Required if prediction_type='raster'.
                                            Contains a list of raster objects containing
                                            some or all fields required to infer the dependent variable value.
        ---------------------------------   -------------------------------------------------------------------------
        datefield                           Optional string. Field name from feature layer
                                            that contains the date, time for the input features.
                                            Same as `prepare_tabulardata()`.
        ---------------------------------   -------------------------------------------------------------------------
        distance_features                   Optional List of Feature Layer objects.
                                            These layers are used for calculation of field "NEAR_DIST_1",
                                            "NEAR_DIST_2" etc in the output dataframe.
                                            These fields contain the nearest feature distance
                                            from the input_features.
                                            Same as `prepare_tabulardata()`.
        ---------------------------------   -------------------------------------------------------------------------
        output_layer_name                   Optional string. Used for publishing the output layer.
        ---------------------------------   -------------------------------------------------------------------------
        gis                                 Optional GIS Object. Used for publishing the item.
                                            If not specified then active gis user is taken.
        ---------------------------------   -------------------------------------------------------------------------
        prediction_type                     Optional String.
                                            Set 'features' or 'dataframe' to make output feature layer predictions.
                                            With this feature_layer argument is required.

                                            Set 'raster', to make prediction raster.
                                            With this rasters must be specified.
        ---------------------------------   -------------------------------------------------------------------------
        output_raster_path                  Optional path.
                                            Required when prediction_type='raster', saves
                                            the output raster to this path.
        ---------------------------------   -------------------------------------------------------------------------
        match_field_names                   Optional dictionary.
                                            Specify mapping of field names from prediction set
                                            to training set.
                                            For example:
                                                {
                                                    "Field_Name_1": "Field_1",
                                                    "Field_Name_2": "Field_2"
                                                }
        =================================   =========================================================================

        :returns Feature Layer if prediction_type='features', dataframe for prediction_type='dataframe' else creates an output raster.

        """

        rasters = explanatory_rasters if explanatory_rasters else []
        if prediction_type in ['features', 'dataframe']:

            if input_features is None:
                raise Exception("Feature Layer required for predict_features=True")

            gis = gis if gis else arcgis.env.active_gis
            return self._predict_features(input_features, rasters, datefield, distance_features, output_layer_name, gis, match_field_names, prediction_type)
        else:
            if not rasters:
                raise Exception("Rasters required for predict_features=False")

            if not output_raster_path:
                raise Exception("Please specify output_raster_folder_path to save the output.")

            return self._predict_rasters(output_raster_path, rasters, match_field_names)

    def _predict_features(
            self,
            input_features,
            rasters=None,
            datefield=None,
            distance_feature_layers=None,
            output_name="Prediction Layer",
            gis=None,
            match_field_names=None,
            prediction_type="features"
    ):
        if isinstance(input_features, FeatureLayer):
            dataframe = input_features.query().sdf
        else:
            dataframe = input_features.copy()

        fields_needed = self._data._categorical_variables + self._data._continuous_variables
        distance_feature_layers = distance_feature_layers if distance_feature_layers else []
        continuous_variables = self._data._continuous_variables

        columns = dataframe.columns
        feature_layer_columns = []
        for column in columns:
            column_name = column
            categorical = False

            if column_name in fields_needed:
                if column_name not in continuous_variables:
                    categorical = True
            elif match_field_names and match_field_names.get(column_name):
                if match_field_names.get(column_name) not in continuous_variables:
                    categorical = True
            else:
                continue

            feature_layer_columns.append((column_name, categorical))

        raster_columns = []
        if rasters:
            for raster in rasters:
                column_name = raster.name
                categorical = False
                if column_name in fields_needed:
                    if column_name not in continuous_variables:
                        categorical = True
                elif match_field_names and match_field_names.get(column_name):
                    column_name = match_field_names.get(column_name)
                    if column_name not in continuous_variables:
                        categorical = True
                else:
                    continue

                raster_columns.append((raster, categorical))

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            processed_dataframe, fields_mapping = TabularDataObject._prepare_dataframe_from_features(
                input_features,
                self._data._dependent_variable,
                feature_layer_columns,
                raster_columns,
                datefield,
                distance_feature_layers
            )

        if match_field_names:
            processed_dataframe.rename(columns=match_field_names, inplace=True)

        for field in fields_needed:
            if field not in processed_dataframe.columns:
                raise Exception(f"Field missing {field}")

        for column in processed_dataframe.columns:
            if column not in fields_needed:
                processed_dataframe = processed_dataframe.drop(column, axis=1)

        processed_numpy = self._data._process_data(processed_dataframe.reindex(sorted(processed_dataframe.columns), axis=1))
        predictions = self._predict(processed_numpy)
        dataframe["prediction_results"] = predictions

        if prediction_type == "dataframe":
            return dataframe

        return dataframe.spatial.to_featurelayer(output_name, gis)

    def _predict_rasters(self, output_folder_path, rasters, match_field_names=None):

        if not os.path.exists(os.path.dirname(output_folder_path)):
            raise Exception("Output directory doesn't exist")

        if os.path.exists(output_folder_path):
            raise Exception("Output Folder already exists")

        try:
            import arcpy
        except:
            raise Exception("This function requires arcpy.")

        try:
            import numpy as np
        except:
            raise Exception("This function requires numpy.")

        try:
            import pandas as pd
        except:
            raise Exception("This function requires pandas.")

        if not HAS_FAST_PROGRESS:
            raise Exception("This function requires fastprogress.")

        fields_needed = self._data._categorical_variables + self._data._continuous_variables

        try:
            arcpy.env.outputCoordinateSystem = rasters[0].extent['spatialReference']['wkt']
        except:
            arcpy.env.outputCoordinateSystem = rasters[0].extent['spatialReference']['wkid']

        xmin = rasters[0].extent['xmin']
        xmax = rasters[0].extent['xmax']
        ymin = rasters[0].extent['ymin']
        ymax = rasters[0].extent['ymax']
        min_cell_size_x = rasters[0].mean_cell_width
        min_cell_size_y = rasters[0].mean_cell_height

        default_sr = rasters[0].extent['spatialReference']

        for raster in rasters:
            point_upper = arcgis.geometry.Point(
                {'x': raster.extent['xmin'], 'y': raster.extent['ymax'], 'sr': raster.extent['spatialReference']})
            point_lower = arcgis.geometry.Point(
                {'x': raster.extent['xmax'], 'y': raster.extent['ymin'], 'sr': raster.extent['spatialReference']})
            cell_size = arcgis.geometry.Point(
                {'x': raster.mean_cell_width, 'y': raster.mean_cell_height, 'sr': raster.extent['spatialReference']})

            points = arcgis.geometry.project([point_upper, point_lower, cell_size], raster.extent['spatialReference'],
                                             default_sr)
            point_upper = points[0]
            point_lower = points[1]
            cell_size = points[2]

            if xmin > point_upper.x:
                xmin = point_upper.x
            if ymax < point_upper.y:
                ymax = point_upper.y
            if xmax < point_lower.x:
                xmax = point_lower.x
            if ymin > point_lower.y:
                ymin = point_lower.y

            if min_cell_size_x > cell_size.x:
                min_cell_size_x = cell_size.x

            if min_cell_size_y > cell_size.y:
                min_cell_size_y = cell_size.y

        max_raster_columns = int(abs(math.ceil((xmax-xmin)/min_cell_size_x)))
        max_raster_rows = int(abs(math.ceil((ymax-ymin)/min_cell_size_y)))

        point_upper = arcgis.geometry.Point({'x': xmin, 'y': ymax, 'sr': default_sr})
        cell_size = arcgis.geometry.Point({'x': min_cell_size_x, 'y': min_cell_size_y, 'sr': default_sr})

        raster_data = {}
        for raster in rasters:
            field_name = raster.name
            point_upper_translated = arcgis.geometry.project([point_upper], default_sr, raster.extent['spatialReference'])[0]
            cell_size_translated = arcgis.geometry.project([cell_size], default_sr, raster.extent['spatialReference'])[0]
            if field_name in fields_needed:
                raster_read = raster.read(origin_coordinate=(point_upper_translated.x, point_upper_translated.y), ncols=max_raster_columns, nrows=max_raster_rows, cell_size=(cell_size_translated.x, cell_size_translated.y))
                for row in range(max_raster_rows):
                    for column in range(max_raster_columns):
                        values = raster_read[row][column]
                        index = 0
                        for value in values:
                            key = field_name
                            if index != 0:
                                key = key + f'_{index}'
                            if not raster_data.get(key):
                                raster_data[key] = []
                            index = index + 1
                            raster_data[key].append(value)
            elif match_field_names and match_field_names.get(raster.name):
                field_name = match_field_names.get(raster.name)
                raster_read = raster.read(origin_coordinate=(point_upper_translated.x, point_upper_translated.y), ncols=max_raster_columns, nrows=max_raster_rows, cell_size=(cell_size_translated.x, cell_size_translated.y))
                for row in range(max_raster_rows):
                    for column in range(max_raster_columns):
                        values = raster_read[row][column]
                        index = 0
                        for value in values:
                            key = field_name
                            if index != 0:
                                key = key + f'_{index}'
                            if not raster_data.get(key):
                                raster_data[key] = []
                            index = index + 1
                            raster_data[key].append(value)
            else:
                continue

        for field in fields_needed:
            if field not in list(raster_data.keys()) and match_field_names and match_field_names.get(field, None) is None:
                raise Exception(f"Field missing {field}")

        processed_data = []

        length_values = len(raster_data[list(raster_data.keys())[0]])
        for i in range(length_values):
            processed_row = []
            for raster_name in sorted(raster_data.keys()):
                processed_row.append(raster_data[raster_name][i])
            processed_data.append(processed_row)

        processed_df = pd.DataFrame(data=np.array(processed_data), columns=sorted(raster_data))

        processed_numpy = self._data._process_data(processed_df)

        predictions = self._predict(processed_numpy)

        predictions = np.array(predictions.reshape([max_raster_rows, max_raster_columns]), dtype='float64')

        processed_raster = arcpy.NumPyArrayToRaster(predictions, arcpy.Point(xmin, ymin), x_cell_size=min_cell_size_x, y_cell_size=min_cell_size_y)
        processed_raster.save(output_folder_path)

        return True
